# -*- coding: utf-8 -*-
"""
_SetState_

Oracle implementation of Locations.SetState
Created on Mon Jun 18 14:09:57 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Locations.SetState import SetState as MySQLSetState

class SetState(MySQLSetState):
    """
    Same as MySQL version
    """
    pass
