#!/usr/bin/env python
"""
_UpdateBlocks_

Oracle implementation of DBS3Buffer.UpdateBlocks
"""

from WMComponent.DBS3Buffer.MySQL.UpdateBlocks import UpdateBlocks as MySQLUpdateBlocks

class UpdateBlocks(MySQLUpdateBlocks):
    """
    Identical to MySQL version.
    """
