"""
_UpdateLocation_

Oracle implementation of Jobs.UpdateLocation

Created on Apr 18, 2013

@author: dballest
"""

from WMCore.WMBS.MySQL.Jobs.UpdateLocation import UpdateLocation as MySQLUpdateLocation

class UpdateLocation(MySQLUpdateLocation):
    pass
