#!/usr/bin/env python
"""
_ListJobStates_

Oracle implementation of Monitoring.ListJobStates
"""




from WMCore.WMBS.MySQL.Monitoring.ListJobStates import ListJobStates \
    as ListJobStatesMySQL

class ListJobStates(ListJobStatesMySQL):
    pass
