"""
_GetPSNtoPNNMapping_

Oracle implementation of Locations.GetPSNtoPNNMapping

"""
from __future__ import (print_function, division)

from WMCore.WMBS.MySQL.Locations.GetPSNtoPNNMapping import GetPSNtoPNNMapping as MySQLGetPSNtoPNNMapping

class GetPSNtoPNNMapping(MySQLGetPSNtoPNNMapping):
    """
    Same as MySQL version

    """
    pass
