#!/usr/bin/env python
"""
_UpdateFiles_

Oracle implementation of DBS3Buffer.UpdateFiles
"""

from WMComponent.DBS3Buffer.MySQL.UpdateFiles import UpdateFiles as MySQLUpdateFiles

class UpdateFiles(MySQLUpdateFiles):
    pass
